#!/usr/bin/env python
import sys, re, subprocess, collections

#####

def allsum_fwd(s):
    return sum(map(ord, s))

ROT_AMOUNT = 7
def rotxor_fwd(s, val=0):
    for c in s:
        val = val ^ ord(c)
        val = (((val << ROT_AMOUNT) | (val >> (32 - ROT_AMOUNT))) & 0xFFFFFFFF)
    return (((val << ROT_AMOUNT) | (val >> (32 - ROT_AMOUNT))) & 0xFFFFFFFF)
def rotxor_bwd(s):
    val = 0
    for c in s[::-1]:
        val = (((val << (32 - ROT_AMOUNT)) | (val >> ROT_AMOUNT)) & 0xFFFFFFFF)
        val = val ^ ord(c)
    assert not rotxor_fwd(s, val)
    return val

DEFAULT_HASH = rotxor_bwd

REQUIRED_SYMBOLS = {
    "kernel32.dll": [
        "ExitProcess",
        "GlobalLock",
        "GlobalUnlock",
        "GetStdHandle",
        "WriteFile",
        "LoadLibraryA",
    ],
    "user32.dll": [
        "OpenClipboard",
        "CloseClipboard",
        "GetClipboardData",
    ],
    "gdi32.dll": [],
    "opengl32.dll": [],
    "advapi32.dll": [],
}

#####

def GetSymbolList(dll):
    pedump = subprocess.Popen(["pedump", "C:\\Windows\\SysWoW64\\" + dll], stdout=subprocess.PIPE)
    for line in pedump.stdout:
        m = re.match(r'\s+[0-9A-F]+\s+\d+\s+([A-Za-z0-9_]+)', line)
        if m:
            yield m.group(1)
    pedump.wait()

def GetHashMap(symlist, hashfunc=DEFAULT_HASH):
    return {sym: hashfunc(sym) for sym in symlist}

def ReverseHashMap(hmap):
    revmap = collections.defaultdict(list)
    for sym, h in hmap.iteritems():
        revmap[h].append(sym)
    return revmap

def GetTotalCollisions(revmap):
    return sum(map(len, revmap.itervalues())) - len(revmap)

def GetCollisionsForSymbol(fwdmap, revmap, sym):
    entry = revmap[fwdmap[sym]]
    assert sym in entry
    return (len(entry) - 1, sum(int(x < sym) for x in entry))

#####

any_coll = 0
critical_coll = 0
for dll in sorted(REQUIRED_SYMBOLS):
    fwdmap = GetHashMap(GetSymbolList(dll))
    revmap = ReverseHashMap(fwdmap)
    tc = GetTotalCollisions(revmap)
    any_coll += tc
    print "%-16s%4d" % (dll, tc)
    for sym in sorted(REQUIRED_SYMBOLS[dll]):
        tc, tb = GetCollisionsForSymbol(fwdmap, revmap, sym)
        print "%-22s%-24s%4d%4d  0x%08X" % ("", sym, tc, tb, fwdmap[sym])
        critical_coll += tb
if critical_coll:
    print "FAILED - Critical Collisions Found."
    sys.exit(1)
elif any_coll:
    print "ACCEPTABLE - No Critical Collisions Found, But No General-Purpose Hash"
else:
    print "PASSED - Good General-Purpose Hash"
